# -*- coding: utf-8 -*-
import base64
import datetime
import hmac
import logging
import requests
import werkzeug

from collections import defaultdict
from dateutil.relativedelta import relativedelta
from hashlib import sha256
from werkzeug.urls import url_encode

from odoo import api, Command, models, fields, _, SUPERUSER_ID
from odoo.osv import expression
from odoo.exceptions import AccessError, UserError, ValidationError
from odoo.addons.iap.tools import iap_tools
from odoo.addons.saas_worker.auth_util import SaasAuth
from odoo.addons.saas_worker.const import WEBSITE, MASTER
from odoo.addons.saas_worker.util import email_on_exception, humanize_version
from odoo.addons.openerp_enterprise.models.spoof import no_impostor
from odoo.addons.openerp_enterprise.models.openerp_enterprise import HOSTING_TYPES_SELECTION

from odoo.addons.sale_subscription.models.sale_order import \
    SUBSCRIPTION_PROGRESS_STATE, SUBSCRIPTION_CLOSED_STATE, SUBSCRIPTION_DRAFT_STATE

_logger = logging.getLogger(__name__)


# Enforced by FME
class SalesRules():

    def _get_sh_product_ids(self, env):
        sh_worker_product_id = env.ref('openerp_enterprise.product_odoo_sh_extra_worker').id
        sh_storage_product_id = env.ref('openerp_enterprise.product_odoo_sh_extra_gb').id
        sh_dedicated_product_id = env.ref('openerp_enterprise.product_odoo_sh_dedicated').id

        return sh_worker_product_id, sh_storage_product_id, sh_dedicated_product_id

    def _get_2022_standard_plan_product_ids(self, env):
        standard_plan_id = env.ref('openerp_enterprise.product_odoo_user_standard').id
        standard_plan_discount_id = env.ref('openerp_enterprise.product_odoo_user_standard_discount').id
        return standard_plan_id, standard_plan_discount_id

    def _get_2022_custom_plan_product_ids(self, env):
        custom_plan_id = env.ref('openerp_enterprise.product_odoo_user_custom').id
        custom_plan_discount_id = env.ref('openerp_enterprise.product_odoo_user_custom_discount').id
        return custom_plan_id, custom_plan_discount_id

    def _get_sh_qty(self, env, lines):

        sh_worker_product_id, sh_storage_product_id, sh_dedicated_product_id = self._get_sh_product_ids(env)

        sh_worker_qty = 0
        sh_storage_qty = 0
        sh_dedicated_qty = 0

        # Each line should be unique per product, but we handle the case anyway
        for line in lines:
            if line['product_id'] == sh_worker_product_id:
                sh_worker_qty += line['qty']
            if line['product_id'] == sh_storage_product_id:
                sh_storage_qty += line['qty']
            if line['product_id'] == sh_dedicated_product_id:
                sh_dedicated_qty += line['qty']

        return (sh_worker_qty, sh_storage_qty, sh_dedicated_qty)

    # Ensure that each product line is unique.
    # Only exception is for the subcription extra discount
    def _check_rule_lines_double_product_lines(self, env, lines):

        excluded_ids = [
            env.ref('openerp_enterprise.product_odoo_subscription_extra_discount').id,
            env.ref('openerp_enterprise.product_odoo_subscription_referral_per_country').id,
            env.ref('openerp_enterprise.product_odoo_subscription_monthly_success_pack_basic').id,
            env.ref('openerp_enterprise.product_odoo_subscription_monthly_success_pack_standard').id,
            env.ref('openerp_enterprise.product_learning_partner_discount').id,
            env.ref('openerp_enterprise.product_official_partner_discount').id,
            env.ref('openerp_enterprise.product_prorated_discount').id,
            env.ref('openerp_enterprise.product_light_user_discount').id,
            env.ref('openerp_enterprise.product_multiyear_discount').id,
        ]

        product_ids = [
            line['product_id']
            for line in lines
            if line['product_id'] not in excluded_ids and line['recurring_invoice']
        ]
        if len(product_ids) > len(set(product_ids)):
            raise UserError(
                "Each Pricing Plan product line should be unique.\n"
                "- if you are upselling, create a new upsell from the main subscription and edit the product quantity\n"
                "- if you want to sell in advance some user, add every user on the same plan line "
                "and prorate the use of them with the \"Discount\" product\n"
                "- Odoo SH product should only have a single line. "
                "If you are upselling, create a new upsell from the main subscription and edit the quantity of the sh line.\n"
                "- You can't have more than 1 Odoo User product line."
            )

    # More than 1 Dedicated server does not make any sense.
    def _check_rule_lines_ensure_one_dedicated(self, env, lines):

        sh_worker_qty, sh_storage_qty, sh_dedicated_qty = self._get_sh_qty(env, lines)

        if sh_dedicated_qty > 1:
            raise UserError("You can't have more than 1 Odoo SH Dedicated Server.")

    # a Dedicated server alone does not make any sense.
    def _check_rule_lines_dedicated_not_alone(self, env, lines):

        sh_worker_qty, sh_storage_qty, sh_dedicated_qty = self._get_sh_qty(env, lines)

        if sh_dedicated_qty == 1 and all(qty <= 0 for qty in [sh_worker_qty, sh_storage_qty]):
            raise UserError("You must have another SH product (Worker or Staging) if you plan to use an SH Dedicated Server.")

    # Check upper limits on workers and storage for Dedicated Servers
    def _check_rule_lines_dedicated_max_limits(self, env, lines):

        sh_worker_qty, sh_storage_qty, sh_dedicated_qty = self._get_sh_qty(env, lines)
        MAX_DEDICATED_WORKERS = 64
        MAX_DEDICATED_STORAGE = 4096

        if sh_dedicated_qty > 0:
            if sh_worker_qty > MAX_DEDICATED_WORKERS:
                raise UserError(
                    f"A dedicated Odoo SH server shouldn't have more than {MAX_DEDICATED_WORKERS} workers."
                )

            if sh_storage_qty > MAX_DEDICATED_STORAGE:
                raise UserError(
                    f"A dedicated Odoo SH server shouldn't use more than {MAX_DEDICATED_STORAGE} GB."
                )

    # Check if a Dedicated Server should be used
    def _check_rule_lines_dedicated_required(self, env, lines):

        sh_worker_qty, sh_storage_qty, sh_dedicated_qty = self._get_sh_qty(env, lines)

        MAX_WORKERS = 8
        MAX_STORAGE = 512

        if sh_worker_qty > MAX_WORKERS and not sh_dedicated_qty:
            raise UserError(
                f"You must use a dedicated Odoo SH Server if you plan to use more than {MAX_WORKERS} workers."
            )

        if sh_storage_qty > MAX_STORAGE and not sh_dedicated_qty:
            raise UserError(
                f"You must use a dedicated Odoo SH Server if you plan to use more than {MAX_STORAGE} GB."
            )

    # You should have storages if you use extra workers
    def _check_rule_lines_storage_required(self, env, lines):

        sh_worker_qty, sh_storage_qty, sh_dedicated_qty = self._get_sh_qty(env, lines)

        if sh_worker_qty > 0 and not sh_storage_qty > 0:
            raise UserError(
                "You should have storage (Odoo.sh GBs) lines if you use Odoo.sh Workers."
            )

    # And vice-versa
    def _check_rule_lines_workers_required(self, env, lines):

        sh_worker_qty, sh_storage_qty, sh_dedicated_qty = self._get_sh_qty(env, lines)

        if sh_storage_qty > 0 and not sh_worker_qty > 0:
            raise UserError(
                "You should have Odoo.sh Workers lines if you use Odoo.sh Storage (Odoo.sh GBs)."
            )

    def _check_rule_lines_upsell_no_recurring(self, env, lines):

        allowed_categories_ids = [
            env.ref('openerp_enterprise.product_category_odoo_users').id,
            env.ref('openerp_enterprise.category_openerp_apps').id,
            env.ref('openerp_enterprise.product_category_odoo_sh').id,
        ]
        allowed_products_ids = [
            env.ref('openerp_enterprise.product_odoo_once_discount').id,
        ]
        non_recurring_upsell_lines = [sl for sl in lines if sl['upsell'] and not sl['recurring_invoice']]

        if any([
            sl['product_categ_id'] not in allowed_categories_ids
            and sl['product_id'] not in allowed_products_ids
            for sl in non_recurring_upsell_lines
        ]):
            raise UserError(
                "You can only upsell recurring products."
            )

    def _check_lines(self, env, lines):

        for name in dir(self):
            if name.startswith('_check_rule_lines_'):
                method = getattr(self, name)
                if callable(method):
                    method(env, lines)

    def _check_rule_so_lifecycle_valid_sub_origin(self, upsell_so):
        origin_sub = upsell_so.origin_order_id

        if origin_sub.exists() and (
           not origin_sub.start_date or
           origin_sub.state != 'sale' or
           origin_sub.subscription_state not in (SUBSCRIPTION_PROGRESS_STATE + SUBSCRIPTION_CLOSED_STATE)
           ):

            raise UserError(
                f"Breaking constraint on {upsell_so.name} [{upsell_so.id}].\n\n"
                f"The original subscription ({ origin_sub.name } [{ origin_sub.id }]) "
                "is either not in a valid state or never started."
            )

    def _check_so_lifecycle_rules(self, record):

        for name in dir(self):
            if name.startswith('_check_rule_so_lifecycle_'):
                method = getattr(self, name)
                if callable(method):
                    method(record)

    def _check_rule_sub_2022_plan_product_mismatch(self, sub):

        standard_product_ids = self._get_2022_standard_plan_product_ids(sub.env)
        custom_product_ids = self._get_2022_custom_plan_product_ids(sub.env)
        standard_sub_products = sub.order_line.product_id.filtered(lambda p: p.id in standard_product_ids)
        custom_sub_products = sub.order_line.product_id.filtered(lambda p: p.id in custom_product_ids)

        if sub.env.context.get('skip_upsell_plan_check'):
            return

        if sub._is_custom_plan() and len(standard_sub_products) > 0:
            p_names = ", ".join(standard_sub_products.mapped('name'))
            raise UserError(
                f"Breaking constraint on {sub.name} [{sub.id}].\n\n"
                f"The product(s) ({ p_names }) cannot be added to a custom plan."
            )

        if sub.is_subscription and sub._is_standard_plan() and len(custom_sub_products) > 0:
            p_names = ", ".join(custom_sub_products.mapped('name'))
            error_msg = f"Breaking constraint on {sub.name} [{sub.id}].\n\n"
            error_msg += f"The product(s) ({p_names}) cannot be added to a standard plan."
            if sub.state == 'sale':
                error_msg += ("\nPlease remove the standard line to save the subscription "
                              "and create an upsell to change the plan.")
            raise UserError(
                error_msg
            )

    def _check_rule_sub_2022_plan_reccurence_mismatch(self, sub):
        if sub.plan_id != sub.sale_order_template_id.plan_id:

            if not sub.sale_order_template_id.plan_id:
                raise UserError(
                    f"Breaking constraint on {sub.name} [{sub.id}].\n\n"
                    f"The plan ({sub.sale_order_template_id.name}) does not have a recurrence."
                    "Please change the plan or the recurrence to save the quote."
                )
            elif not sub.plan_id:
                raise UserError(
                    f"Breaking constraint on {sub.name} [{sub.id}].\n\n"
                    f"The subscription ({sub.name}) does not have a recurrence."
                    "Please change the plan or the recurrence to save the quote."
                )
            else:
                raise UserError(
                    f"Breaking constraint on {sub.name} [{sub.id}].\n\n"
                    f"The recurrence ({sub.plan_id.name}) of the subscription ({sub.name})"
                    f"does not match the plan ({sub.sale_order_template_id.name})."
                    "Please change the plan or the recurrence to save the quote."
                )

    def _check_rule_sub_2022_archived_product(self, sub):
        archived_products = sub.order_line.product_id.filtered(lambda p: p.active is False)
        if len(archived_products) > 0:
            raise UserError(
                f"Breaking constraint on {sub.name} [{sub.id}].\n\n"
                "You cannot confirm a quote with archived products. "
                "Restart your upsell from the parent subscription "
                "and edit the quantity on the order line."
            )

    def _check_rule_sub_2022_commission_zero_percent_discount(self, sub):

        discount_lp_tmpl = sub.env.ref('openerp_enterprise.product_learning_partner_discount_product_template')
        discount_op_tmpl = sub.env.ref('openerp_enterprise.product_official_partner_discount_product_template')
        discount_products = (discount_lp_tmpl + discount_op_tmpl).mapped('product_variant_id')

        zero_percent_comm = sub.env.ref('openerp_portal.zero_percent_commission_to_partner')

        if sub.commercial_partner_id.id == sub.referrer_id.commercial_partner_id.id and \
           sub.commission_plan_id.id != zero_percent_comm.id:

            if any([p.id in discount_products.ids for p in sub.order_line.mapped('product_id')]):
                raise UserError(
                    f"Breaking constraint on {sub.name} [{sub.id}].\n\n"
                    "You cannot have a partner discount product unless the subscription has a 0%% commission plan."
                )

    def _check_rule_sub_2022_no_lone_maintenance_line(self, sub):
        standard_plan = sub.env.ref('openerp_enterprise.product_odoo_user_standard')
        std_line_qty = sum(sub.order_line.filtered(
            lambda line: line.product_id == standard_plan
        ).mapped('product_uom_qty'))
        # In the middle of a confirmation of switch plan uspsell, we skip the check
        if sub._context.get('skip_upsell_plan_check') and not std_line_qty:
            return

        if sub._get_maintenance_line() \
           and not (sub._is_custom_plan() or sub.maintenance_main_subscription_id):

            raise UserError(
                f"Breaking constraint on {sub.name} [{sub.id}].\n\n"
                "Any maintenance line should be on a subscription with either "
                "a custom plan or a separate main subscription on which the maintenance applies."
            )

    def _check_rule_sub_no_1styear_discount_after_its_expiration(self, sub):
        user_discount_products = (
            sub.env.ref('openerp_enterprise.product_odoo_user_discount'),
            sub.env.ref('openerp_enterprise.product_odoo_user_standard_discount'),
            sub.env.ref('openerp_enterprise.product_odoo_user_custom_discount'),
        )
        has_first_year_discount = any(
            line for line in sub.order_line
            if line.product_id in user_discount_products and line.product_uom_qty > 0
        )
        if (
            sub.state in ('sent', 'sale') and sub.subscription_id
            and has_first_year_discount and sub.first_contract_date
        ):
            first_year_disc_end_date = sub.first_contract_date + relativedelta(years=1, days=-1)
            start_date = sub.start_date  # upsell date or start/renewal date
            if start_date and start_date > first_year_disc_end_date:
                raise UserError(
                    f"Breaking constraint on {sub.name} [{sub.id}].\n\n"
                    "You cannot have user initial discount products past 1 year after the first contract date."
                )

    def _check_sub_rules(self, record):
        for name in dir(self):
            if name.startswith('_check_rule_sub_'):
                method = getattr(self, name)
                if callable(method):
                    method(record)

    def _check_so_default(self, sale_order):

        standardized_lines = self._standardize_order_lines(sale_order.order_line)
        self._check_lines(sale_order.env, standardized_lines)

    def _check_so_upsell(self, sale_order):

        self._check_so_lifecycle_rules(sale_order)

        standardized_lines = self._standardize_order_lines(sale_order.order_line, upsell=True)
        standardized_parent_lines = self._standardize_order_lines(sale_order.subscription_id.order_line)
        parent_lines_by_id = {l['id']: l for l in standardized_parent_lines}

        for line in standardized_lines:
            # Sum lines with their parent if any
            if line['parent_line_id'] and line['parent_line_id'] in parent_lines_by_id.keys():
                line['qty'] += parent_lines_by_id[line['parent_line_id']]['qty']
                del parent_lines_by_id[line['parent_line_id']]

        # Add the remaining parent lines
        standardized_lines.extend(parent_lines_by_id.values())

        self._check_lines(sale_order.env, standardized_lines)



    def _check_so(self, sale_order):

        if sale_order.subscription_state == "7_upsell":
            self._check_so_upsell(sale_order)
        else:
            self._check_so_default(sale_order)

    def _check_sub(self, subscription):

        self._check_sub_rules(subscription)
        self._check_so_lifecycle_rules(subscription)
        standardized_lines = self._standardize_order_lines(subscription.order_line)
        self._check_lines(subscription.env, standardized_lines)

    def _standardize_order_lines(self, so_lines, upsell=False):

        standardized_lines = []

        for line in so_lines:
            if line['display_type'] not in ('line_section', 'line_note'):
                standardized_lines.append({
                    'id': line.id,
                    'product_id': line.product_id.id,
                    'product_categ_id': line.product_id.categ_id.id,
                    'qty': line.product_uom_qty,
                    'parent_line_id': line.parent_line_id.id if line.parent_line_id else False,
                    'recurring_invoice': line.recurring_invoice,
                    'upsell': upsell,
                })

        return standardized_lines


class SaleSubscription(models.Model):
    _inherit = "sale.order"
    _mail_post_access = 'read'  # allow posting without write access
    _check_company_auto = False
    _rec_names_search = ['name', 'client_order_ref']

    country_id = fields.Many2one('res.country', related='partner_id.country_id', store=True)
    industry_id = fields.Many2one('res.partner.industry', related='partner_id.industry_id', store=True)
    client_order_ref = fields.Char(tracking=True)

    enterprise_final_customer_id = fields.Many2one('res.partner', 'Final Customer', tracking=True, index=True)
    enterprise_saas = fields.Boolean('Is Hosted by Odoo', compute='_compute_enterprise_saas', store=True)
    enterprise_users = fields.Integer(compute="_compute_enterprise_users", store=True, tracking=True)
    database_ids = fields.One2many('openerp.enterprise.database', 'subscription_id', 'Databases')
    database_count = fields.Integer('Databases #', compute='_compute_database_count')
    enterprise_has_download = fields.Boolean("Has Access to Enterprise Download", compute="_compute_has_download", search="_search_has_download")
    enterprise_is_partnership = fields.Boolean("Is a valid Partnership subscription", compute='_is_partnership', search='_search_is_partnership')
    enterprise_is_enterprise = fields.Boolean("Is a valid Enterprise subscription", compute='_is_enterprise', search="_search_is_enteprise")
    responsible_id = fields.Many2one('res.users', 'Responsible', tracking=True)
    enterprise_security_email = fields.Char(
        'Security Contact', help="Optional. If set, this list of comma-separated emails will receive "
        "Security Advisories instead of the default contacts (customer + integrating partner)", tracking=True)

    subscription_tag_ids = fields.Many2many(
        comodel_name='openerp.enterprise.subscription.tag',
        string='Subscription Tags',
        domain="['|', ('company_id', '=', False), ('company_id', '=', company_id)]",
        check_company=True,
    )

    odoo_sh_enterprise_subscription_id = fields.Many2one(
        'sale.order',
        string="Odoo.sh Enterprise Subscription",
        index='btree_not_null',
        tracking=True,
        help="Enterprise subscription for this Odoo.sh subscription"
             " (only when Odoo Users are on a separated subscription)"
    )
    odoo_sh_subscription_id = fields.One2many(
        'sale.order', 'odoo_sh_enterprise_subscription_id',
        string="Odoo.sh Subscription",
        readonly=True,
        help="Odoo.sh subscription linked to this enterprise contract"
             " (only when Odoo SH products are on a separated subscription)",
    )
    odoo_sh_extra_gb = fields.Integer('Odoo.sh Extra GB', tracking=True)
    odoo_sh_extra_stagings = fields.Integer('Odoo.sh Extra Stagings', tracking=True)
    odoo_sh_extra_workers = fields.Integer('Odoo.sh Extra Workers', tracking=True)
    odoo_sh_extra_dedicated = fields.Integer('Odoo.sh Extra Dedicated', tracking=True)
    has_odoo_sh = fields.Boolean('Has Odoo.sh products', compute='_compute_has_odoo_sh')

    pricelist_id = fields.Many2one('product.pricelist', 'Pricelist', tracking=True)
    next_invoice_date = fields.Date(tracking=True)

    hosting = fields.Selection(HOSTING_TYPES_SELECTION,
                               compute='_compute_hosting', inverse='_inverse_hosting',
                               store=True, recursive=True, index=True, tracking=True)
    version = fields.Char(compute="_compute_version", store=True, index='btree_not_null')

    # track to renew field
    to_renew = fields.Boolean(compute="_compute_to_renew", search="_search_to_renew", tracking=True)

    exclude_dashboard = fields.Boolean('Exclude User Change', tracking=True,
        help="Exclude future user change from partner dashboard")

    is_staff_restricted = fields.Boolean(
        string="Staff Restriction", tracking=True, default=False,
        help="When enabled, a customer authorization is required for each database access)",
    )
    spoof_authorization_ids = fields.One2many('openerp.enterprise.spoof.authorization', 'subscription_id', 'Staff Authorizations')
    expected_next_invoice_date = fields.Date(
        string="Expected",
        compute='_compute_expected_next_invoice_date',
        help="Theorical next invoice date, based on start date, recurrence and current time")
    invoice_date_diff = fields.Integer(compute='_compute_expected_next_invoice_date')
    company_id = fields.Many2one(tracking=True)

    plan_billing_period_unit = fields.Selection(string="Plan Billing Period",
                                                related='plan_id.billing_period_unit', readonly=True)
    warning_subscription_id = fields.Many2one(
        'sale.order',
        string="Possible Duplicate Subscription",
        compute="_compute_warning_subscription_id",
    )

    @api.onchange('referrer_id')
    def _onchange_referrer(self):
        self.hosting = False  # reset hosting to force user/salesman to set the correct one

    @api.depends('subscription_state')
    def _compute_warning_subscription_id(self):
        for sub in self:
            if sub.subscription_state in ['3_progress', '4_paused']:
                domain = [
                    ('enterprise_is_enterprise', '=', sub.enterprise_is_enterprise),
                    ('enterprise_is_partnership', '=', sub.enterprise_is_partnership),
                    ('subscription_state', 'in', ['3_progress', '4_paused']),
                    ('id', '!=', sub.id)
                ]
                if sub.database_ids:
                    domain = expression.AND([domain, [('database_ids', '=', False)]])
                existing_duplicate = self._check_duplicate(sub.commercial_partner_id.email or sub.partner_id.email, domain)[:1]
                sub.warning_subscription_id = existing_duplicate
            else:
                sub.warning_subscription_id = False

    @api.model
    def _check_duplicate(self, email, domain):
        current_email_domain_criterion = iap_tools.mail_prepare_for_domain_search(
            email
        )
        if not current_email_domain_criterion:
            return self.env['sale.order']
        partner_ids = self.env['crm.lead'].search([
            ('email_domain_criterion', '=', current_email_domain_criterion)
        ]).mapped('partner_id')
        search_domain = [
            ('partner_id', 'child_of', partner_ids.ids),
            ('is_subscription', '=', True),
        ]
        search_domain = expression.AND((
            search_domain,
            domain,
        ))
        existing_subscription_ids = self.env['sale.order'].search(
            search_domain,
            order="subscription_state",
        )
        return existing_subscription_ids

    @api.depends('next_invoice_date', 'payment_token_id', 'upsell_date', 'subscription_state')
    def _compute_to_renew(self):
        for sub in self:
            if sub.subscription_state in SUBSCRIPTION_DRAFT_STATE:
                sub.to_renew = False
                continue
            expired = (
                not sub.payment_token_id
                and sub.next_invoice_date
                and sub.next_invoice_date <= fields.Date.today()
            )
            sub.to_renew = expired or sub.upsell_date

    @api.depends('database_ids.version')
    def _compute_version(self):
        for so in self:
            so.version = humanize_version(so.database_ids[:1].version or '') or False

    def _search_to_renew(self, operator, value):
        domain = ['&', ('subscription_state', 'not in', SUBSCRIPTION_DRAFT_STATE),
                  '|', ('upsell_date', '!=', False),
                         '&', ("next_invoice_date", "<=", fields.Date.today()),
                              ('payment_token_id', '=', False)]
        if operator in expression.NEGATIVE_TERM_OPERATORS:
            value = not value
        if value:
            return domain
        if not value:
            return ['!'] + domain

    @api.depends('start_date', 'next_invoice_date')
    def _compute_expected_next_invoice_date(self):
        # dynamic value - cannot be stored!
        today = fields.Date.today()
        for sub in self:
            if (not sub.is_subscription or not sub.start_date or not sub.plan_id
                    or (sub.plan_id.billing_period_unit not in ('month', 'year'))
                    or not sub.plan_id.billing_period_value):
                sub.expected_next_invoice_date = False
                sub.invoice_date_diff = 0
                continue
            # no support for recurrence in days/weeks
            recurrence_months = (sub.plan_id.billing_period_value * (12 if sub.plan_id.billing_period_unit == 'year' else 1))
            elapsed_recurrences = max(int((today - sub.start_date).days / recurrence_months / (365.2425 / 12)), 0)
            next_recurrence = elapsed_recurrences + 1
            sub.expected_next_invoice_date = sub.start_date + relativedelta(months=next_recurrence * recurrence_months)
            if sub.next_invoice_date:
                sub.invoice_date_diff = abs((sub.next_invoice_date - sub.expected_next_invoice_date).days)
            else:
                sub.invoice_date_diff = 0

    def _track_subtype(self, init_values):
        self.ensure_one()
        if 'to_renew' in init_values:
            return self.env.ref('sale_subscription.subtype_state_change')
        return super(SaleSubscription, self)._track_subtype(init_values)

    def _set_subscription_code(self):
        def is_valid_number(contract_number):
            return contract_number and len(contract_number) >= 13 and contract_number.startswith("M")

        for sub in self.filtered(lambda s: s.is_subscription and not is_valid_number(s.client_order_ref)):
            sub.client_order_ref = self.env['ir.sequence'].with_company(sub.company_id).next_by_code('sale.subscription')

    @api.model_create_multi
    def create(self, vals_list):
        res = super(SaleSubscription, self).create(vals_list)

        res._set_subscription_code()
        return res

    def _move_info_to_renewing_subscription(self):
        for sub in self.sudo():
            # Move the database on the renewal subscription at confirmation
            sub.subscription_id.database_ids.subscription_id = sub
            sub.is_staff_restricted = sub.subscription_id.is_staff_restricted
            # Move github user
            sub.write({
                'github_user_ids': [Command.set(sub.subscription_id.github_user_ids.ids)]
            })
            # Move tasks
            tasks = self.env['project.task'].sudo().search([('mnt_subscription_id', '=', sub.subscription_id.id)])
            tasks.mnt_subscription_id = sub
            # Move link between enterprise sur and odoo.sh sub
            # When odoo.sh subscription is renewed
            sub.subscription_id.odoo_sh_enterprise_subscription_id.odoo_sh_subscription_id = sub
            # When the main subscription is renewed
            sub.odoo_sh_subscription_id = sub.subscription_id.odoo_sh_subscription_id

    def _confirm_renewal(self):
        super()._confirm_renewal()
        self.env.cr.precommit.add(self._move_info_to_renewing_subscription)

    @api.multi
    def write(self, vals):
        # we forbid to move a subscription to the "close" stage, instead the user has to use the close wizard
        if not self.env.su and not self.env.context.get('close_wizard'):
            if self and vals.get('subscription_state') in ['6_churn']:
                raise ValidationError("""
                    You cannot move a subscription to a `close` state.
                    If you wish to close a subscription use its close wizard on the form view.
                """)

        # We seperate the active and inactive subscriptions from self to check
        # if we need to monitor the changes on the subscriptions lines.
        closed_subscriptions = self.filtered(lambda s: s.subscription_state not in ['3_progress', '4_paused'])
        running_subscriptions = self - closed_subscriptions

        if 'subscription_state' in vals:
            new_state = vals['subscription_state']
            if (closed_subscriptions and new_state in ('3_progress', '4_paused')) or (running_subscriptions and new_state not in ('3_progress', '4_paused')):

                old_values = self._get_tracked_sh_values()
                res = super(SaleSubscription, self).write(vals)
                new_values = self._get_tracked_sh_values()
                self.env['sale.order.line']._monitor_tracked_sh_product_changes(old_values, new_values)

                return res

        res = super(SaleSubscription, self).write(vals)

        if 'company_id' in vals:
            # force taxes to be recomputation as superuser to ensure taxes from
            # previous company get correctly removed. (we need to invalidate cache
            # beforehand, so that we reload all values)
            order_lines = self.order_line
            order_lines.invalidate_recordset(['tax_id'], flush=False)
            order_lines.sudo()._compute_tax_id()

            # force the warehouse corresponding to the new company
            # assuming that we only have one warehouse per company here at Odoo
            self.warehouse_id = self.env.user.with_company(vals['company_id'])._get_default_warehouse_id()

        if 'sale_order_template_id' in vals:
            self._set_subscription_code()

        return res

    def copy(self, default=None):
        # To remove when the duplication behavior is fixed properly in standard module
        if self.filtered(lambda s: s.subscription_state in ['7_upsell', '2_renewal', '3_progress']) \
                and not self.env.su and not self.env.context.get('force_copy_subscription'):
            raise UserError("You cannot duplicate a running subscription, a renewal or an upsell. "
                            "Create a quote from scratch if you want a new subscription. "
                            "Click on copy to X if you want an alternative duration for the same customer. "
                            "Cancel the current upsell and upsell again if you cannot change the current upsell.")
        return super().copy(default=default)

    @api.model
    def _name_search(self, name, domain=None, operator='ilike', limit=None, order=None):
        if name and operator in ('like', 'ilike'):
            domain = list(domain or ())
            enterprise_code, _sep, order_name = name.partition('-')
            if enterprise_code.strip() and order_name.strip():
                # matching subscription display name, search for records
                # matching enterprise_code AND order_name
                domain = expression.AND((
                    domain,
                    [
                        ('client_order_ref', operator, enterprise_code.strip()),
                        ('name', operator, order_name.strip())
                    ],
                ))
                return self._search(domain, limit=limit, order=order)
        return super()._name_search(name, domain=domain, operator=operator, limit=limit, order=order)

    def _compute_display_name(self):
        subscriptions = self.filtered(lambda x: x.is_subscription and x.client_order_ref)
        orders = self - subscriptions
        if orders:
            super(SaleSubscription, orders)._compute_display_name()

        for record in subscriptions:
            record.display_name = f'{record.client_order_ref} - {record.name}'

    @api.constrains('is_staff_restricted', 'sale_order_template_id')
    def _check_is_staff_restricted(self):
        partner_template = self.env.ref('openerp_enterprise.quotation_template_odoo_partnership')
        for sub in self:
            if sub.is_staff_restricted and sub.sale_order_template_id.id == partner_template.id:
                raise ValidationError("Staff Restriction mode only allowed on Enterprise subscriptions")

    def _notify_change(self):
        """ Build and send the mail with the monitored changes"""
        subscription_values = self.env.cr.precommit.data.pop('sale.subscription.odoosh_modification', {})
        ids = [id_ for id_, vals in subscription_values.items() if vals]
        if not ids:
            return

        ICP = self.env['ir.config_parameter'].sudo()
        nb_workers_trigger_limit = int(ICP.get_param('openerp_enterprise.paas_extra_workers_limit', default=8))
        paas_template = self.env.ref("openerp_enterprise.subscription_paas_warning")
        perf_pack_template = self.env.ref("openerp_enterprise.promote_perf_pack")
        for subscription in self.browse(ids).exists():
            values = subscription_values[subscription.id]
            old_nb_workers = values.get('old_nb_workers', 0)
            new_nb_workers = values.get('new_nb_workers', 0)
            old_nb_dedicated = values.get('old_nb_dedicated', 0)
            new_nb_dedicated = values.get('new_nb_dedicated', 0)
            old_state = values.get('old_state', False)
            new_state = values.get('new_state', False)

            mail_context = {
                'old_nb_workers': old_nb_workers,
                'new_nb_workers': new_nb_workers,
                'old_nb_dedicated': old_nb_dedicated,
                'new_nb_dedicated': new_nb_dedicated,
                'old_state': old_state,
                'new_state': new_state
            }
            if (
                (old_state != new_state and new_state == '3_progress' and new_nb_workers > 0)
                or
                (subscription.subscription_state == '3_progress' and old_nb_workers == 0 and new_nb_workers > 0)
            ):
                subscription.with_context(mail_context).message_post_with_source(
                    perf_pack_template,
                    subtype_xmlid='mail.mt_comment',
                )

            # Check if the sub is in progress or move from/to progress
            progress = '3_progress' in (old_state, new_state) or subscription.subscription_state == '3_progress'
            # Check if product change need to be notified
            notify_product_change = (
                max(old_nb_workers, new_nb_workers) > nb_workers_trigger_limit or
                old_nb_dedicated != new_nb_dedicated
            )
            # Do notify les 2 peys qui rapportent des thunes a.k.a. the PaaS Team if needed
            if progress and notify_product_change:
                paas_template.with_context(mail_context).send_mail(subscription.id)

    @api.constrains('odoo_sh_enterprise_subscription_id', 'order_line')
    def _check_odoosh_enterprise_subscription(self):
        for sub in self:
            if sub.odoo_sh_enterprise_subscription_id:
                if sub.is_subscription and sub.has_odoo_sh and sub._get_billed_user_count():
                    raise ValidationError(
                        "You can't both specify an Odoo.sh Enterprise subscription "
                        "and have Odoo Users on the same subscription"
                    )
                if sub.odoo_sh_enterprise_subscription_id.odoo_sh_subscription_id - sub:
                    raise ValidationError(
                        "The choosen Odoo.sh Enterprise Subscription is already linked "
                        "to another Odoo.sh Subscription"
                    )

    def _get_tracked_sh_products(self):
        """ Returns the products that must be monitored by
        the Odoo SH/PaaS Team on the subscriptions
        """
        return {
            'sh_dedicated': self.env.ref('openerp_enterprise.product_odoo_sh_dedicated'),
            'sh_extra_worker': self.env.ref('openerp_enterprise.product_odoo_sh_extra_worker')
        }


    def _get_tracked_sh_values(self, track_stage=False):

        tracked_sh_products = self._get_tracked_sh_products()
        values = defaultdict(dict)
        for subscription in self:
            for sh_product_code in tracked_sh_products:
                sub_line_ids = subscription.order_line

                product_line_ids = sub_line_ids.filtered(lambda sl: sl.product_id == tracked_sh_products[sh_product_code])
                values[subscription.id][sh_product_code] = sum(product_line_ids.mapped('product_uom_qty'))
                values[subscription.id]['subscription_state'] = subscription.subscription_state

        return values

    def _mail_track(self, tracked_fields, initial):
        before_logs = self.order_log_ids
        updated_fields, commands = super()._mail_track(tracked_fields, initial)

        old_state = initial.get('subscription_state', self.subscription_state)
        new_state = self.subscription_state
        reopened_order = old_state in ['6_churn', '5_renewed'] and new_state in SUBSCRIPTION_PROGRESS_STATE
        total = sum(self.order_log_ids.mapped('new_enterprise_user'))
        if 'enterprise_users' in updated_fields:
            field_user_count_id = self.env['ir.model.fields']._get(self._name, 'enterprise_users').id
            user_command = next(filter(lambda c: c[2]['field_id'] == field_user_count_id, commands))
            new_user_count = user_command[2]['new_value_integer'] - user_command[2]['old_value_integer']
            after_logs = self.order_log_ids
            last_log = after_logs - before_logs
            renewal_order = old_state == '2_renewal' and new_state in SUBSCRIPTION_PROGRESS_STATE
            if (not last_log and new_state == '5_renewed') or renewal_order:
                # upon renewal confirmation, transfer log are immediately created
                # for parent subscription and it won't be created while tracking that
                # parent subscription, so reuse that pre-existing transfer log
                # In addition the renewal order can create contraction of expansion log and positive user count
                # can end up on a contraction log which is not consistent. In that case force the user count
                # on the transfer log
                last_log = after_logs.sorted().filtered(
                    lambda l: l.event_type in ('3_transfer', '2_churn') and l.new_enterprise_user == 0
                )[:1]

            # If a log has been created, use it
            if last_log:
                # When the upsell reopen the order, the new user count is not correct
                # Since the sub was close and reopen, the new_user_count = enterprise user of the sub
                # But the churn log has been deleted, so we need to calculate to add only the upsell user
                # Actually any use case where mrr changed during the churn phase.
                if reopened_order:
                    new_user_count -= total

                last_log.sorted()[0].write({
                    'referrer_id': self.referrer_id.id,
                    'new_enterprise_user': new_user_count,
                    'include_dashboard': not self.exclude_dashboard,
                })
            # no new log, because no change in mrr: duplicate the last log
            elif self.order_log_ids and not reopened_order:
                self.order_log_ids.sorted()[0].copy(default={
                    'event_type': '1_expansion' if new_user_count > 0 else '15_contraction',
                    'amount_signed': 0,
                    'event_date': fields.Date.today(),
                    'referrer_id': self.referrer_id.id,
                    'new_enterprise_user': new_user_count,
                    'include_dashboard': not self.exclude_dashboard,
                })

        # Case where churn log are deleted and lost users are removed
        # We need to add them back.
        if old_state == '6_churn' and new_state == '5_renewed':
            t_logs = self.order_log_ids.sorted().filtered(lambda l: l.event_type == '3_transfer')
            if total and t_logs:
                t_logs[0].write({
                    'referrer_id': self.referrer_id.id,
                    'new_enterprise_user': -total + t_logs[0].new_enterprise_user,
                    'include_dashboard': not self.exclude_dashboard,
                })

        return updated_fields, commands

    def _get_most_valid_subscription(self, code):
        """
            Return the subscription that is closer to a valid subscription

            We exclude locked subscription since it means there is a confirmed renewal
            in progress on which the database should be linked

            Return an empty record set if there is no subscription with the given code
        """
        return self.search([
            ('client_order_ref', '=ilike', code),
            ('is_subscription', '=', True),
            ('state', 'in', ('draft', 'sent', 'sale')),
            ('locked', '=', False),
        ]).sorted(
            key=lambda s: {
                '3_progress': 0,
                '4_paused': 1,
                '1_draft': 2,
                '2_renewal': 3,
                '6_churn': 4
            }.get(s.subscription_state, 10)
        )[:1]

    def _get_most_valid_subscription_by_codes(self, codes_list):
        """ Return the subscription that is closer to a valid subscription for a list of subscription codes

        We exclude locked subscription since it means there is a confirmed renewal
        in progress on which the database should be linked

        :param codes_list: list or tuple of subscription code
        :return: a dictionnary with subscription code as key and subscription record as value
           (if no valid subscription was found for a code, that code won't be included in the result)
        """
        subscription_by_code = {}
        subscriptions = self.search([
            ('client_order_ref', 'in', codes_list),
            ('is_subscription', '=', True),
            ('state', 'in',  ('draft', 'sent', 'sale')),
            ('locked', '=', False),
        ]).sorted(
            key=lambda s: {
                '3_progress': 0,
                '4_paused': 1,
                '1_draft': 2,
                '2_renewal': 3,
                '6_churn': 4
            }.get(s.subscription_state, 10)
        )
        for subscription in subscriptions:
            subscription_code = subscription.client_order_ref
            if subscription_code in subscription_by_code:
                continue
            subscription_by_code[subscription_code] = subscription
        return subscription_by_code

    @no_impostor
    def ask_for_subscription_access(self):
        self.ensure_one()
        if not self.is_staff_restricted:
            raise UserError('You cannot request access to a subscription with no staff restrictions')
        composer_form_view_id = self.env.ref('mail.email_compose_message_wizard_form').id
        template_id = self.env.ref('openerp_enterprise.subscription_request_access')
        auth_vals = {
            'subscription_id': self.id,
            'reason': '',
            'user_id': self.env.user.id,
            'state': 'draft',
            'responsible_partner_id': self.partner_id.id,
        }
        spoof_auth = self.env['openerp.enterprise.spoof.authorization'].sudo().create(auth_vals)
        request_url = WEBSITE + '/my/subscription/access-form?' + url_encode({
            "id": spoof_auth.id,
            'token': spoof_auth._get_authorization_token()
        })
        ctx = {
            'default_composition_mode': 'comment',
            'default_res_ids': [self.id],
            'default_model': 'sale.order',
            'default_template_id': template_id.id,
            'default_partner_ids': [self.partner_id.id],
            'active_ids': self.ids,
            'task': None,
            'sub_access_request': True,
            'request_url': request_url,
            'spoof_id': spoof_auth.id,
        }
        return {
            'type': 'ir.actions.act_window',
            'view_mode': 'form',
            'res_model': 'mail.compose.message',
            'view_id': composer_form_view_id,
            'target': 'new',
            'context': ctx,
        }

    @api.returns('mail.message', lambda value: value.id)
    def message_post(self, **kwargs):
        if self.env.context.get('sub_access_request'):
            self = self.with_context(mail_create_nosubscribe=True)
            database = self.database_ids[0]
            spoof_vals = {
                'user_id': self.env.user.id,
                'db_uuid': database.db_uuid,
                'db_id': database.id,
                'reason': '',
                'action': "access_request",
                'success': True,
                'restricted': False,  # No restrictions to ask database access
            }
            # Body is already computed so, changing the request url in the context does not update it.
            self.env['openerp.enterprise.spoof.authorization'].sudo().browse(self.env.context.get('spoof_id')).write({'state': 'pending'})
            self.env['openerp.enterprise.spoof'].sudo().create(spoof_vals)
            self._message_log(body="Access requested to subscription")
        return super(SaleSubscription, self).message_post(**kwargs)

    def _notify_by_email_add_values(self, base_mail_values):
        if self.env.context.get('sub_access_request'):
            base_mail_values.update(is_notification=False)
        return super(SaleSubscription, self)._notify_by_email_add_values(base_mail_values)

    @api.depends('order_line')
    def _compute_has_odoo_sh(self):
        odoo_sh_product_ids = [
            self.env.ref('openerp_enterprise.product_odoo_sh').id,
            self.env.ref('openerp_enterprise.product_odoo_sh_extra_worker').id,
        ]

        for r in self:
            r.has_odoo_sh = bool(set(r.order_line.mapped('product_id').ids) & set(odoo_sh_product_ids))

    def action_open_sh_subscription(self):
        self.ensure_one()
        res = self.env['ir.actions.act_window']._for_xml_id('openerp_enterprise.action_oe_sh_subscription')
        res['res_id'] = self.odoo_sh_enterprise_subscription_id.id or self.odoo_sh_subscription_id.id
        return res

    # Computed fields Methods

    @api.depends('database_ids', 'order_line.product_id', 'sale_order_template_id', 'subscription_id.hosting',
                 'subscription_state', 'odoo_sh_subscription_id')
    def _compute_hosting(self):
        product_odoo_sh_worker = self.env.ref('openerp_enterprise.product_odoo_sh_extra_worker', raise_if_not_found=False)
        partnership_tag = self.env.ref('openerp_enterprise.odoo_partnership_tag', raise_if_not_found=False)
        for subscription in self:
            if partnership_tag and partnership_tag in self.sale_order_template_id.subscription_tag_ids:
                # Partnership has no related hosting
                subscription.hosting = ''
            elif any([line.product_id == product_odoo_sh_worker for line in subscription.order_line]) \
                    or subscription.odoo_sh_subscription_id:
                # if this subscription has Odoo.sh, then it host by Odoo.sh (pretty obvious, no?)
                subscription.hosting = 'paas'
            elif subscription._is_standard_plan():
                subscription.hosting = 'saas'
            elif subscription.subscription_state == '7_upsell':
                # upsell take the same hosting as its parent
                subscription.hosting = subscription.subscription_id.hosting
            else:
                # otherwise use the pre-existing value (if set by user/salesman)
                # or (if existing) the related database hosting
                # or fallback to 'saas' by default
                subscription.hosting = (
                    subscription.hosting
                    or subscription.database_ids[:1].hosting
                    or 'saas'
                )

    def _inverse_hosting(self):
        # force recompute to ensure user did not set an invalid value
        self._compute_hosting()

    # NOTE : ideally, database_ids.info should be in dependencies, since it will chagne the result of is_saas().
    # For performance issue, I guess we should not be that to avoid a computation at each ping.

    @api.depends('database_ids')
    def _compute_enterprise_saas(self):
        for subscription in self:
            subscription.enterprise_saas = any(db.is_saas() for db in subscription.database_ids)

    @api.depends('subscription_state', 'order_line.product_id', 'order_line.product_uom_qty')
    def _compute_enterprise_users(self):
        user_products = self.env.ref('openerp_enterprise.product_odoo_user_custom', raise_if_not_found=False)
        if not user_products:
            return
        user_products |= self.env.ref('openerp_enterprise.product_odoo_user_standard')
        for subscription in self:
            user_lines = subscription.order_line.filtered(lambda l: l.product_id in user_products)
            if subscription.subscription_state in ['3_progress', '4_paused']:
                subscription.enterprise_users = sum(user_lines.mapped("product_uom_qty"))
            else:
                subscription.enterprise_users = 0


    def _compute_database_count(self):
        database_data = self.env['openerp.enterprise.database'].read_group([('subscription_id', 'in', self.ids)], ['subscription_id'], ['subscription_id'])
        mapped_data = dict([(db['subscription_id'][0], db['subscription_id_count']) for db in database_data])
        for subscription in self:
            subscription.database_count = mapped_data.get(subscription.id, 0)


    def _compute_has_download(self):
        """ Check if the contract should have access to the Enterprise version downloads """
        part_tag_id = self.env.ref('openerp_enterprise.odoo_partnership_tag')
        for contract in self:
            partnership = contract.sale_order_template_id.subscription_tag_ids & part_tag_id
            yearly_custom_plan = (
                contract._is_custom_plan() and
                contract.plan_id.billing_period_unit == 'year'
            )
            contract.enterprise_has_download = (
                contract.subscription_state in ['3_progress', '4_paused'] and
                (partnership or yearly_custom_plan)
            )

    def _search_has_download(self, operator, value):
        tag_ids = [
            self.env.ref('openerp_enterprise.odoo_partnership_tag').id,
        ]

        custom_tag = self.env.ref('openerp_enterprise.custom_plan_tag')

        if (operator, value) in [('=', True), ('!=', False)]:
            return [
                '&',
                ('subscription_state', 'in', ['3_progress', '4_paused']),
                '|', ('sale_order_template_id.subscription_tag_ids', 'in', tag_ids),
                          '&', ('sale_order_template_id.subscription_tag_ids', 'in', custom_tag.ids),  # noqa: E131
                               ('plan_id.billing_period_unit', '=', 'year'),
            ]
        return [
            '|',
            ('subscription_state', 'not in', ['3_progress', '4_paused']),
            '&', ('sale_order_template_id.subscription_tag_ids', 'not in', tag_ids),
                      '|', ('sale_order_template_id.subscription_tag_ids', 'not in', custom_tag.ids),  # noqa: E131
                           ('plan_id.billing_period_unit', '!=', 'year'),
        ]

    def _search_is_partnership(self, operator, value):
        partnership_tag = self.env.ref('openerp_enterprise.odoo_partnership_tag')

        if (operator, value) in [('=', True), ('!=', False)]:
            return [('subscription_state', 'in', ('3_progress', '4_paused')),
                    ('sale_order_template_id.subscription_tag_ids', '=', partnership_tag.id)]

        return ['|', ('subscription_state', 'not in', ('3_progress', '4_paused')),
                     ('sale_order_template_id.subscription_tag_ids', '!=', partnership_tag.id)]


    def _is_partnership(self):
        partnership_tag = self.env.ref('openerp_enterprise.odoo_partnership_tag')
        for contract in self:
            contract.enterprise_is_partnership = (contract.subscription_state in ('3_progress', '4_paused') and
                                                  partnership_tag in contract.sale_order_template_id.sudo().subscription_tag_ids)

    def _is_enterprise(self):
        for subscription in self:
            subscription.enterprise_is_enterprise = subscription._is_enterprise_plans() and \
                                                  subscription.subscription_state in ('3_progress', '4_paused')

    def _is_standard_plan(self):
        standard_plan = self.env.ref('openerp_enterprise.standard_plan_tag', raise_if_not_found=False)
        return standard_plan and standard_plan in self.sale_order_template_id.subscription_tag_ids

    def _is_custom_plan(self):
        custom_plan = self.env.ref('openerp_enterprise.custom_plan_tag', raise_if_not_found=False)
        return custom_plan and custom_plan in self.sale_order_template_id.subscription_tag_ids

    def _is_enterprise_plans(self):
        return self._is_standard_plan() or self._is_custom_plan()

    def _search_is_enteprise(self, operator, value):
        plan_tags = self.env.ref('openerp_enterprise.standard_plan_tag')
        plan_tags |= self.env.ref('openerp_enterprise.custom_plan_tag')
        if (operator, value) in [('=', True), ('!=', False)]:
            return [
                ('subscription_state', 'in', ('3_progress', '4_paused')),
                ('sale_order_template_id.subscription_tag_ids', 'in', plan_tags.ids),
            ]
        return ['|',
            ('subscription_state', 'not in', ('3_progress', '4_paused')),
            ('sale_order_template_id.subscription_tag_ids', 'not in', plan_tags.ids),
        ]

    # Business Methods

    def _prepare_apps_product_odoo_sh(self):
        res = []
        # dict to map product
        odoo_sh_products = {
            'odoo_sh_extra_gb': self.env.ref('openerp_enterprise.product_odoo_sh_extra_gb'),
            'odoo_sh_extra_stagings': self.env.ref('openerp_enterprise.product_odoo_sh_extra_staging'),
            'odoo_sh_extra_workers': self.env.ref('openerp_enterprise.product_odoo_sh_extra_worker'),
            'odoo_sh_extra_dedicated': self.env.ref('openerp_enterprise.product_odoo_sh_dedicated')
        }

        # This dict is used to translate to Odoo.sh Access product to the new odoo.sh products
        odoo_sh_access_to_extra = {
            'odoo_sh_extra_gb': 16,
            'odoo_sh_extra_stagings': 3,
            'odoo_sh_extra_workers': 2,
            'odoo_sh_extra_dedicated': 0,
        }
        odoo_sh_access_product_qty = sum(self.order_line.filtered(lambda x:
                                            x.product_id == self.env.ref('openerp_enterprise.product_odoo_sh')).mapped('product_uom_qty'))

        for k in odoo_sh_products:
            # compute how much odoo workers, gb, stagings this subscription needs
            qty_needed = self[k] - odoo_sh_access_to_extra[k] * odoo_sh_access_product_qty
            # find if a line is already present for this product
            line = self.order_line.filtered(lambda x: x.product_id == odoo_sh_products[k])[:1]
            # if at least one line exists, we update it
            if line:
                if qty_needed > 0:
                    # update the quantity
                    values = {
                        'product_uom_qty': qty_needed,
                    }
                    res.append((1, line.id, values))
                # if we do not need anymore this product we delete the line only if it is an automatic one
                elif line.computation == 'automatic':
                    res.append((2, line.id))
            # if no lines found we create it
            else:
                # we create the line only if the quantity needed is positive
                if not qty_needed > 0:
                    continue
                product = odoo_sh_products[k]
                # compute the line name
                name = product.display_name
                if product.description_sale:
                    name += '\n' + product.description_sale
                # retrieve the uom we should use for this subscription template
                uom_id = self.sale_order_template_id.get_product_uom(product)
                price = product.with_context(pricelist=self.pricelist_id.id, uom=uom_id.id)._get_contextual_price()
                values = {
                    'name': name,
                    'computation': 'automatic',
                    'product_id': product.id,
                    'product_uom_qty': qty_needed,
                    'price_unit': price,
                    'product_uom': uom_id.id,
                }
                res.append((0, 0, values))


        return res

    def _get_invoicing_company(self):
        self.ensure_one()
        foreign_payment_companies = (
            self.env.ref('openerp_enterprise.company_dmcc'),
            self.env.ref('openerp_enterprise.company_ke'),
            self.env.ref('openerp_enterprise.company_idn')
        )
        if self.company_id in foreign_payment_companies:
            pm_company_id = self.company_id
        else:
            pm_company_id = self.payment_token_id and self.payment_token_id.provider_id.sudo().company_id
        return pm_company_id or self.company_id

    def _prepare_invoice(self):
        """ Force the company depending on the payment token (Odoo Ltd invoicing in Odoo S.A.).

        Also override standard values for recurring invoices:
        - Set the SaaS salesman and salesteam on monthly invoices if needed
        - Force partner's payment terms
        """
        self.ensure_one()

        invoice_company = self._get_invoicing_company()
        self = self.with_company(invoice_company)
        result = super()._prepare_invoice()
        if invoice_company != self.company_id:
            result.update({
                'fiscal_position_id': self.fiscal_position_id._get_fiscal_position(self.partner_invoice_id).id,
                'partner_bank_id': invoice_company.partner_id.bank_ids.filtered(lambda bank: bank.company_id.id in (invoice_company.id, False))[:1].id,
                'company_id': invoice_company.id,
            })

        if self.is_subscription:
            if self.plan_id.billing_period_unit == "month":
                saas_user = self.env.ref('openerp_enterprise.oe_saas_user')
                saas_team = self.env.ref('openerp_enterprise.be_online_team')
                if self.user_id.id == SUPERUSER_ID or not self.user_id:
                    result['user_id'] = saas_user.id
                    result['team_id'] = saas_team.id
                else:
                    result['user_id'] = self.user_id.id
                    result['team_id'] = self.user_id.sale_team_id.id or saas_team.id

        return result

    def _get_subscription_mail_payment_context(self, mail_ctx=None):
        email_context = super()._get_subscription_mail_payment_context(mail_ctx=mail_ctx)
        # force error email to be sent to online@openerp.com
        email_context["responsible_email"] = "online@openerp.com"
        return email_context

    def _create_invoices(self, grouped=False, final=False, date=None):
        if self and not self.env.su:
            # check that user have access to the company into which the subscription
            # invoice will be created, if not warn it beforehand trying to create
            # the real invoice.
            for subscription in self.filtered('is_subscription'):
                invoice_company = subscription._get_invoicing_company().sudo()
                if invoice_company not in self.env.companies:
                    raise UserError(_(
                        "Invoices for subscription '%s' will be created in "
                        "company '%s' but you don't have access to that company.",
                        subscription.display_name, invoice_company.name,
                    ))
        return super()._create_invoices(grouped=grouped, final=final, date=date)

    @api.model
    @email_on_exception()
    def _cron_recurring_create_invoice(self):
        return super(SaleSubscription, self)._cron_recurring_create_invoice()

    def _filter_odoo_monthly_contract(self):
        def is_odoo_monthly_contract(contract):
            return contract._is_enterprise_plans() \
                and contract.plan_id.billing_period_unit != 'year'

        return self.filtered(is_odoo_monthly_contract)

    def _filter_subscription_to_invoice(self, automatic=False):
        current_date = fields.Date.today()

        def close_yearly_contract(contract):
            return contract.plan_id.billing_period_unit == 'year' \
                and contract.upsell_date and contract.upsell_date < current_date

        # close subscriptions that need to be closed
        close_contract_ids = self.filtered(close_yearly_contract)
        close_contract_ids.set_close()
        # update the products on our odoo monthly subscriptions
        monthly_odoo_subscription = self._filter_odoo_monthly_contract()
        monthly_odoo_subscription.set_apps_recurring_lines()
        monthly_odoo_subscription._send_payment_reminder()

        # do not process closed subscription
        self = self - close_contract_ids
        # If automatic mode do not process subscription without token
        if automatic:
            self = self.filtered(lambda s: (
                s.payment_token_id or
                # Don't filter out free subscription
                (s._invoice_is_considered_free(s._get_invoiceable_lines())[0] and s.state == 'sale')
            ))
        return self

    def _get_default_payment_link_values(self):
        # ensure to update the products on our odoo monthly subscriptions
        # before generating the payment link
        if self.subscription_state != '2_renewal':
            monthly_odoo_subscription = self._filter_odoo_monthly_contract()
            monthly_odoo_subscription.set_apps_recurring_lines()
        return super()._get_default_payment_link_values()

    def _send_payment_reminder(self):
        today = fields.Date.today()
        to_remind_subs = self.filtered(lambda s: (
            not s.payment_token_id and
            # filter out free subscription
            not s._invoice_is_considered_free(s._get_invoiceable_lines())[0] and
            # Do not send an email everyday
            (today - s.next_invoice_date).days in [0, 3, 7, (s.plan_id.auto_close_limit or 15) - 1]
        ))
        to_remind_subs.message_post_with_source('openerp_enterprise.payment_reminder_email')

    def _archive_token_if_card_deactivated(self, tx):
        """Archive token if transaction error indicate the card expired/was stolen."""
        self.ensure_one()
        if tx and tx.state == 'error' and tx.ogone_error_code and tx.token_id == self.payment_token_id:
            payment_token = tx.token_id
            error_type = tx._ogone_payment_error_type()
            if error_type in ('card-expired', 'card-stolen'):
                _logger.info('archiving and removing subscription token %s, reason: %s ', payment_token, error_type)
                payment_token_link = payment_token._get_html_link()
                payment_token.active = False
                self.payment_token_id = False
                messages = {
                    "card-expired": _("The Payment Token %(link)s has been removed from the subscription and archived"
                                      " because the related card has expired"),
                    "card-stolen": _("The Payment Token %(link)s has been removed from the subscription and archived"
                                     " because the related card is considered stolen"),
                }
                self.message_post(body=messages[error_type] % {'link': payment_token_link})

    def _handle_subscription_payment_failure(self, invoice, transaction):
        super()._handle_subscription_payment_failure(invoice, transaction)
        self._archive_token_if_card_deactivated(transaction)

    def _create_recurring_invoice(self, batch_size=30):
        # since saas-16.1, this method is only called from cron (so implicit automatic=True)
        self = self._filter_subscription_to_invoice(automatic=True)
        if self:
            return super()._create_recurring_invoice(batch_size=batch_size)
        return self.env['account.move']

    def _recurring_invoice_get_ids(self, company_id=None, modulo=None, mod_remainder=0):
        """ Method to return the subscription ids for the renewal cron
            This method is used to split the subscriptions renewal
            on multiple crons and speed it up.

            Parameters:
                company_id (`res.company`/`None`): Filter by companies, None for no filtering.
                modulo (`int`/`None`): Do a modulo on the subscription id and keep the subscription if the result is the value of `mod_remainder`.
                                        None for no filtering.
                mod_remainder (`int`): The remainder that the modulo operation need to match to keep the subscription id

            Returns:
                list(int): a list of subscription ids matching the filtering, empty list if nothing is found.

        """
        extra_domain = []
        if company_id:
            extra_domain = expression.AND([
                extra_domain,
                [('company_id', 'in', company_id.ids)]
            ])

        search_domain = self._recurring_invoice_domain(extra_domain=extra_domain)

        # search for yearly subscriptions that should be closed because of an expired upsell date
        upsell_domain = [
            ('upsell_date', '<', fields.Date.today()),
            ('plan_id.billing_period_unit', '=', 'year'),
            ('subscription_state', 'in', ['3_progress', '4_paused']),
        ]
        if company_id:
            upsell_domain.append(('company_id', 'in', company_id.ids))
        search_domain = expression.OR([
            search_domain,
            upsell_domain,
        ])

        record_ids = self.search_fetch(search_domain, []).ids
        if modulo:
            return [rid for rid in record_ids if (rid % modulo) == mod_remainder]
        return record_ids

    @email_on_exception()
    def _cron_subscription_expiration(self):
        last_month = fields.Date.subtract(fields.Date.today(), days=30)
        domain_unsuspend = [
            ('is_subscription', '=', True),
            ('next_invoice_date', '<', last_month),
            ('subscription_state', '=', '4_paused'),
        ]
        subscription_to_unsuspend = self.search(domain_unsuspend)
        subscription_to_unsuspend.subscription_state = '3_progress'
        return super()._cron_subscription_expiration()

    def need_odoo_sh_upsell(self):
        self.ensure_one()

        if self.subscription_state != '3_progress':
            return False

        odoo_sh_extra_products = {
            'odoo_sh_extra_gb': self.env.ref('openerp_enterprise.product_odoo_sh_extra_gb').sudo().id,
            'odoo_sh_extra_stagings': self.env.ref('openerp_enterprise.product_odoo_sh_extra_staging').sudo().id,
            'odoo_sh_extra_workers': self.env.ref('openerp_enterprise.product_odoo_sh_extra_worker').sudo().id,
            'odoo_sh_extra_dedicated': self.env.ref('openerp_enterprise.product_odoo_sh_dedicated').sudo().id,
        }
        # This dict is used to translate to Odoo.sh Access product to the new odoo.sh products
        odoo_sh_access_to_extra = {
            'odoo_sh_extra_gb': 16,
            'odoo_sh_extra_stagings': 3,
            'odoo_sh_extra_workers': 2,
            'odoo_sh_extra_dedicated': 0,
        }

        odoo_sh_access_product = self.env.ref('openerp_enterprise.product_odoo_sh').sudo()
        odoo_sh_access_lines_qty = sum(self.order_line.filtered(lambda x: x.product_id.id == odoo_sh_access_product.id)
                                                      .mapped('product_uom_qty'))

        for k, v in odoo_sh_extra_products.items():
            extra_product_qty = sum(self.order_line.filtered(lambda x: x.product_id.id == v).mapped('product_uom_qty'))
            extra_product_qty += max(0, odoo_sh_access_lines_qty) * odoo_sh_access_to_extra[k]
            if self[k] > 0 and self[k] > extra_product_qty:
                return True
        return False

    def _get_odoo_sh_invoice_diff(self):
        self.ensure_one()
        # 3. Odoo.sh extras
        odoo_sh_access_product = self.env.ref('openerp_enterprise.product_odoo_sh').sudo()
        odoo_sh_extra_products = {
            'odoo_sh_extra_gb': self.env.ref('openerp_enterprise.product_odoo_sh_extra_gb').sudo(),
            'odoo_sh_extra_stagings': self.env.ref('openerp_enterprise.product_odoo_sh_extra_staging').sudo(),
            'odoo_sh_extra_workers': self.env.ref('openerp_enterprise.product_odoo_sh_extra_worker').sudo(),
            'odoo_sh_extra_dedicated': self.env.ref('openerp_enterprise.product_odoo_sh_dedicated').sudo(),
        }

        odoo_sh_access_to_extra = {
            'odoo_sh_extra_gb': 16,
            'odoo_sh_extra_stagings': 3,
            'odoo_sh_extra_workers': 2,
            'odoo_sh_extra_dedicated': 0,
        }
        odoo_sh_access_lines_qty = sum(self.order_line
            .filtered(lambda x: x.product_id.id == odoo_sh_access_product.id)
            .mapped('product_uom_qty')
        )
        current_lines = self.upsell_order_id.order_line
        current_extra_lines = current_lines.filtered(lambda sol: sol.product_id.id in [p.id for p in list(odoo_sh_extra_products.values())])
        upsell = []
        for k, v in odoo_sh_extra_products.items():
            extra_product_qty = sum(self.order_line.filtered(lambda x: x.product_id == v).mapped('product_uom_qty'))
            extra_product_qty += max(0, odoo_sh_access_lines_qty) * odoo_sh_access_to_extra[k]
            extra_product_qty_in_so = sum(current_extra_lines.filtered(lambda x: x.product_id == v).mapped('product_uom_qty'))  # the qty on the upsell SO
            if self[k] > 0 and self[k] > extra_product_qty + extra_product_qty_in_so:
                upsell.append({
                    'product_uom_qty': self[k] - (extra_product_qty + extra_product_qty_in_so),
                    'product_id': v.id,
                    'product_uom': self.sale_order_template_id.get_product_uom(v).id,
                    'name': v.display_name,
                })
        return upsell

    def enterprise_upsell(self):
        """
        Generate an upsell quote or fetch the existing one if applicable.

        Check if an existing upsell quote is still applicable; if not, unlink it
        and generate a new one.

        :return: id of the upsell quote (sale.order)
        :rtype: int
        """
        self.ensure_one()

        # check if current upsell is still good or must be re-generated
        diff = self._get_invoice_diff()[self.id]
        if self.upsell_order_id:
            if diff:
                # cancel current upsell if we need a new one
                if self.upsell_order_id.state == 'sent':
                    old_order = self.upsell_order_id.with_context(maintain_upsell_date=True)
                    _logger.info("Cancelled obsolete upsell sale.order #%s for subscription #%s" %
                                 (old_order.id, self.id))
                    old_order.action_cancel()
                    diff = self._get_invoice_diff()[self.id]
                else:
                    raise UserError('Only "Sent" Upsell orders should be set on a subscription.')
            else:
                # if not, send old upsell quote
                return self.upsell_order_id.id
        if not diff:
            return False
        # let's use the existing wizard (to avoid side-porting fixes & behaviours)
        # upsell date is 1 month after extra usage, pro-rata should be computed for that date
        self.upsell_date = self.upsell_date or fields.Date.to_string(datetime.date.today()
                                                                     + relativedelta(months=1))
        init_date = fields.Date.from_string(self.upsell_date) - relativedelta(months=1)

        lang = self.partner_id.lang or self.env.user.lang
        upsell_msg_body = self._get_order_digest(origin='upsell', lang=lang)
        action = self._prepare_renew_upsell_order("7_upsell", upsell_msg_body)
        upsell_order = self.browse(action['res_id'])
        upsell_order.order_line.filtered(lambda l: l.product_uom_qty == 0).unlink()
        upsell_order.write({
            'state': 'sent',
            'require_payment': 1,
            'tag_ids': [(4, self.env.ref('openerp_enterprise.oe_upsell_tag').id, 0)],
            'validity_date': self.upsell_date,
            'date_order': init_date,
        })
        self.write({
            'upsell_order_id': upsell_order.id,
            'to_renew': True,
        })
        return upsell_order.id

    @api.model
    @email_on_exception()
    def _cron_upsell_mail(self):
        """
        Warn yearly customers about pending upsell on detection day and 1 week before expiration.
        """
        Subscription = self.env['sale.order']
        next_month = datetime.date.today() + relativedelta(months=1, days=-1)  # all upsells detected the day before
        next_week = datetime.date.today() + relativedelta(weeks=1)
        sub_mail = Subscription.search([('is_subscription', '=', True),
                                        '|',
                                        ('upsell_date', '=', fields.Date.to_string(next_month)),
                                        ('upsell_date', '=', fields.Date.to_string(next_week))])
        upsell_template = self.env.ref('openerp_enterprise.openerp_enterprise_auto_upsell')
        for sub in sub_mail:
            upsell_template.send_mail(sub.id)
        return True

    def _action_cancel(self):
        # Can only cancel draft/sent quotation, if a db is linked, it come from upgrade-online
        # and should be unlinked
        self.database_ids = False
        return super()._action_cancel()

    def _send_unlink_mail(self):
        database = self.database_ids
        try:
            template = self.env.ref("openerp_enterprise.unlink_already_linked_database")
            mail_context = {
                'email_to': self.partner_id.commercial_partner_id.email,
                'database': database,
                'unlink_url': database._get_unlink_url(),
            }
            template.with_context(mail_context).send_mail(self.id)
            _logger.info("Mail to unlink database %s sent to %s", database.db_name, self.partner_id.commercial_partner_id.email)

        except Exception:
            _logger.exception("Error when sending mail to unlink database %s (%s)", database.db_name, database.db_uuid)

    def _get_portal_url(self, access_token=False):
        # maybe 'sale.subscription' should inherit 'portal.mixin' to access this already existing method
        self.ensure_one()
        return "%s/my/subscription/%d/%s" % (WEBSITE, self.id, self.access_token or '' if access_token else '')

    def _get_unlink_mail_token(self):
        self.ensure_one()
        # hmac key
        db_secret = self.env['ir.config_parameter'].sudo().get_param('database.secret')
        # hmac mesage
        message = '\\%d<>%s/' % (self.id, self.access_token)
        h = hmac.new(db_secret.encode(), digestmod=sha256)
        h.update(message.encode())
        return base64.b64encode(h.digest()).decode()

    def _check_unlink_mail_token(self, token):
        self.ensure_one()
        # compare_digest is a time constant comparison method that takes two bytes var and returns if they are equals
        return hmac.compare_digest(self._get_unlink_mail_token().encode(), token.encode())

    def _get_unlink_mail_send_url(self, json=True):
        self.ensure_one()
        token_quoted = werkzeug.urls.url_quote(self._get_unlink_mail_token())
        if json:
            return "%s/publisher-warranty/send-unlink-mail/json/%d/%s" % (WEBSITE, self.id, token_quoted)
        return "%s/publisher-warranty/send-unlink-mail/%d/%s" % (WEBSITE, self.id, token_quoted)


    def _iot_x509_ids(self):
        """
        Compute certificate identifier based on subscription's contract code,
        skipping subscriptions without an IoT Box.
        """
        iot_box_product = self.env.ref('openerp_enterprise.product_iot_box')
        domain = [
            ('subscription_state', 'in', SUBSCRIPTION_PROGRESS_STATE),
            ('order_line.product_id', '=', iot_box_product.id)
        ]
        if self.ids:
            domain += [('id', 'in', self.ids)]
        res = []
        for sub in self.search(domain):
            res += [sha256(sub.client_order_ref.encode()).hexdigest()[:8]]
        return res

    @api.model
    def _cron_iot_x509(self):
        """
        Compute and send the list of certificate identifiers to MASTER.
        """
        x509_ids = self._iot_x509_ids()
        url = MASTER + '/saas_master/x509/iot/identifiers'
        requests.post(url, json={'params': {'x509_ids': x509_ids}}, auth=SaasAuth())


    def _get_billed_user_count(self):
        """
        Return the billed quantity of user on the subscription
        """
        self.ensure_one()
        if self._is_standard_plan():
            usr_product = self.env.ref('openerp_enterprise.product_odoo_user_standard')
        elif self._is_custom_plan():
            usr_product = self.env.ref('openerp_enterprise.product_odoo_user_custom')
        else:
            # not a user plan (ex. partnership)
            return 0

        usr_line = self.order_line.filtered(lambda l: l.product_id == usr_product)
        return sum(usr_line.mapped('product_uom_qty'))

    def _get_billed_iot_box_count(self):
        """
        Return the billed Iot Box sub. service of the subscription
        """
        self.ensure_one()
        iot_box_product = self.env.ref('openerp_enterprise.product_iot_box')
        iot_box_lines = self.order_line.filtered(lambda l: l.product_id == iot_box_product)
        return sum(iot_box_lines.mapped('product_uom_qty'))

    @api.model
    def _prepare_contract_lines(self, database, partner, pricelist):
        uom_user_month = self.env.ref('openerp_enterprise.product_online_user_month').id
        uom_month = self.env.ref('openerp_enterprise.product_uom_month').id
        if database.get_plan() == 'standard':
            user = self.env.ref('openerp_enterprise.product_odoo_user_standard')
            user_discount = self.env.ref('openerp_enterprise.product_odoo_user_standard_discount')
        else:
            user = self.env.ref('openerp_enterprise.product_odoo_user_custom')
            user_discount = self.env.ref('openerp_enterprise.product_odoo_user_custom_discount')
        default_products = user + user_discount

        lines = []
        for product in default_products:
            qty = database.quantity

            name = product.display_name
            if product.description_sale:
                name += '\n' + product.description_sale
            line = {
                'name': name,
                'product_id': product.id,
                'product_uom_qty': qty,
                'price_unit': product._get_contextual_price(),
                'product_uom': uom_user_month if product in default_products else uom_month,
                'computation': 'automatic',
            }
            lines.append((0, 0, line))
        return lines

    @api.model
    def create_odoo_online_contract(self, vals):
        # retrieve the openerp.enterprise.database record from the vals values
        database_id = self.env['openerp.enterprise.database'].sudo().search([('db_uuid', '=', vals['db_uuid'])])
        # retrieve the customer's res.user record
        customer_user_id = self.env['res.users'].browse(vals['uid'])
        # retrieve the Self Subscription tag
        self_sub_tag_id = self.env.ref('openerp_enterprise.self_subscription_tag')
        # retrieve the pricelist selected for this customer
        pricelist_id = self.env['product.pricelist'].browse(vals['pricelist_id'])
        # compute the subscription lines
        order_line = self._prepare_contract_lines(database_id, customer_user_id.partner_id, pricelist_id)

        # check if the database already is linked to a subscription and if it's a real draft one
        subscription_to_cancel = self.env["sale.order"]
        if (
            database_id.subscription_id and
            database_id.subscription_id.subscription_state in ['1_draft', '2_renewal'] and
            database_id.subscription_id.invoice_count == 0  # make sure it's never been billed
        ):
            # if so we unlink the subscription, mark current sub as 'to be cancelled'
            subscription_to_cancel = database_id.subscription_id

        if database_id.get_plan() == 'standard':
            template = self.env.ref('openerp_enterprise.standard_template_odoo_monthly')
        else:
            template = self.env.ref('openerp_enterprise.custom_template_odoo_monthly')

        # create the subscription
        subscription_id = self.env['sale.order'].sudo().create({
            'origin': 'SaaS Subscription %s' % customer_user_id.partner_id.name,
            'partner_id': customer_user_id.partner_id.id,
            'user_id': SUPERUSER_ID,
            'start_date': datetime.datetime.now(),
            'subscription_state': '1_draft',
            'subscription_tag_ids': [(4, self_sub_tag_id.id, False)],
            # 11MO template
            'sale_order_template_id': template.id,
            'pricelist_id': pricelist_id.id,
            'order_line': order_line,
            # link the database to the subscription.
            'database_ids': [(6, 0, database_id.ids)],
        })
        # if there were a subscription to cancel, do it now
        if subscription_to_cancel:
            subscription_to_cancel.with_context({'disable_cancel_warning': True}).action_cancel()
        # for freemium subscription, we set the subscription in the "in progress" stage directly
        if not subscription_id.amount_total:
            # Thanks to https://github.com/odoo/odoo/commit/19e2e1502742b506a56501468374e4faca9e8334
            # sale logs are not generated properly when the confirm happened in the same transaction
            # than the so creation. We need to clear the precommit to restart log generation
            # Note: in this specific case, precommit stage only contain calls to _track_finalize()
            # so it's fine to call clear()
            self.env.flush_all()
            self.env.cr.precommit.clear()
            subscription_id.action_confirm()

        return subscription_id.id


class SaleSubscriptionLine(models.Model):
    """ Override C(R)UD operations on the subscriptions lines to monitor
    changes related to Odoo SH products that needs to be notified to
    the PaaS Team.
    """
    _inherit = "sale.order.line"

    computation = fields.Selection([('manual', 'Manual'), ('automatic', 'Automatic')], string='Computation', default='manual')
    database_id = fields.Many2one('openerp.enterprise.database', string='Database')



    @api.model_create_multi
    def create(self, vals_list):
        subscription_ids = self.env['sale.order'].browse([vals['order_id'] for vals in vals_list])

        old_values = subscription_ids._get_tracked_sh_values()
        new_lines = super(SaleSubscriptionLine, self).create(vals_list)
        new_values = subscription_ids._get_tracked_sh_values()

        self._monitor_tracked_sh_product_changes(old_values, new_values)

        return new_lines

    @api.multi
    def write(self, vals):
        subscription_ids = self.mapped('order_id')

        old_values = subscription_ids._get_tracked_sh_values()
        res = super(SaleSubscriptionLine, self).write(vals)
        new_values = subscription_ids._get_tracked_sh_values()

        self._monitor_tracked_sh_product_changes(old_values, new_values)

        return res

    @api.multi
    def unlink(self):
        subscription_ids = self.mapped('order_id')

        old_values = subscription_ids._get_tracked_sh_values()
        res = super(SaleSubscriptionLine, self).unlink()
        new_values = subscription_ids._get_tracked_sh_values()

        self._monitor_tracked_sh_product_changes(old_values, new_values)

        return res

    def _prepare_invoice_line(self, **optional_values):
        self.ensure_one()

        invoice_company = self.order_id._get_invoicing_company()
        if invoice_company != self.company_id:
            self = self.with_company(invoice_company)

        res = super()._prepare_invoice_line(**optional_values)

        if invoice_company != self.company_id:
            # when the invoicing company (i.e company from the payment token)
            # differs from the company of the subscription, we need to recompute
            # the taxes according to the target (invoicing) company
            order = self.order_id.with_company(invoice_company)
            if self.product_id:
                fiscal_position = order.fiscal_position_id._get_fiscal_position(order.partner_invoice_id)
                taxes = self.product_id.taxes_id.filtered(lambda t: t.company_id == invoice_company)
                if fiscal_position:
                    taxes = fiscal_position.map_tax(taxes)
                res['tax_ids'] = [(6, 0, taxes.ids)]

        if self.database_id:
            res['name'] += '\n(%s)' % self.database_id.db_name

        return res

    def _monitor_tracked_sh_product_changes(self, old_values, new_values):
        """ Make the mail to the PaaS team with the relevant data (if any)"""

        # Browse all unique subscriptions
        subcriptions_ids = old_values.keys() | new_values.keys()
        subscriptions = self.env['sale.order'].browse(subcriptions_ids)

        precommit_values = self.env.cr.precommit.data.setdefault('sale.subscription.odoosh_modification', {})
        self.env.cr.precommit.add(subscriptions._notify_change)
        for subscription in subscriptions:
            subscription_values = precommit_values.setdefault(subscription.id, {})
            old_subscription_vals = old_values.get(subscription.id, {})
            new_subscription_vals = new_values.get(subscription.id, {})

            # Get the changes that must be notified
            to_notify_values = self._prepare_notify_values(old_subscription_vals, new_subscription_vals)
            # Keep old values if already existing
            for key, value in to_notify_values.items():
                if key.startswith('old'):
                    subscription_values[key] = subscription_values.get(key, value)
                else:
                    subscription_values[key] = value

    def _prepare_notify_values(self, old_subscription_vals, new_subscription_vals):
        """ Return the changes that needs to be notified or
        an empty dict it there are none.
        """
        to_notify_values = {}

        # Get changes based on the keys defined in 'self._get_tracked_sh_products()'
        old_nb_workers = old_subscription_vals.get('sh_extra_worker', 0)
        new_nb_workers = new_subscription_vals.get('sh_extra_worker', 0)
        old_nb_dedicated = old_subscription_vals.get('sh_dedicated', 0)
        new_nb_dedicated = new_subscription_vals.get('sh_dedicated', 0)

        # We also check the stage of the subscription to see if we really need to notify the changes
        old_state = old_subscription_vals.get('subscription_state')
        new_state = new_subscription_vals.get('subscription_state')
        subscription_state_change = False

        # If the subscription was deactived, we notify the stage change
        # and every relevant Odoo.sh products quantities as 0.
        if old_state == '3_progress' and new_state != '3_progress':
            new_nb_dedicated = 0
            new_nb_workers = 0
            subscription_state_change = True

        # If the subscription became in progress, we notify the stage change
        # and every relevant Odoo.sh products quantities as new quantities.
        if old_state != '3_progress' and new_state == '3_progress':
            old_nb_dedicated = 0
            old_nb_workers = 0
            subscription_state_change = True

        if old_nb_dedicated != new_nb_dedicated:
            to_notify_values.update({
                'old_nb_dedicated': old_nb_dedicated,
                'new_nb_dedicated': new_nb_dedicated
            })

        if new_nb_workers != old_nb_workers:
            to_notify_values.update({
                'old_nb_workers': old_nb_workers,
                'new_nb_workers': new_nb_workers,
            })

        if subscription_state_change:
            to_notify_values.update({
                'old_state': old_state,
                'new_state': new_state,
            })

        return to_notify_values


class SaleSubscriptionLog(models.Model):
    _inherit = 'sale.order.log'

    new_enterprise_user = fields.Integer()
    referrer_id = fields.Many2one('res.partner', index=True)
    include_dashboard = fields.Boolean('Included in Partner Dashboard')
    event_date = fields.Date(index=True)
    subscription_code = fields.Char(related="order_id.client_order_ref", store=True)
    # TODO number of loc

    def action_toggle_include_dashboard(self):
        if not self.env.user.has_group('sales_team.group_sale_manager'):
            raise AccessError(_("You cannot perform this action"))

        for rec in self.sudo():
            rec.include_dashboard = not rec.include_dashboard

class SaleSubscriptionLogReport(models.Model):
    _inherit = "sale.order.log.report"

    new_enterprise_user = fields.Integer()
    include_dashboard = fields.Boolean('Included in Partner Dashboard')
    subscription_code = fields.Char()
    hosting = fields.Selection(HOSTING_TYPES_SELECTION, string="Hosting")

    def _select(self):
        # We want to use the more granular information we have directly on the log for the referrer
        # Since new user gain before the referrer was referrer doesn't count
        return super()._select().replace('so.referrer_id', 'log.referrer_id') + """
            ,
            log.new_enterprise_user,
            log.include_dashboard,
            log.subscription_code,
            so.hosting
        """

    def _group_by(self):
        return super()._group_by() + """
            ,
            log.new_enterprise_user,
            log.include_dashboard,
            log.subscription_code,
            so.hosting
        """

    def action_toggle_include_dashboard(self):
        # id is id of the order log
        return self.env["sale.order.log"].browse(self.ids).action_toggle_include_dashboard()


class SaleSubscriptionCloseReason(models.Model):
    _inherit = 'sale.order.close.reason'

    external = fields.Boolean("External",
        help="Subscription close with this reason will not be taken into account for the computation of the retention rate")


class SaleSubscriptionCloseReasonWizard(models.TransientModel):
    _inherit = "sale.subscription.close.reason.wizard"


    def set_close(self):
        # we add a new key in the context to warn the code that we close using the wizard and allow
        # the current user to set the subscription's stage to close.
        user = self.env.user
        sudo = user.has_group('sales_team.group_sale_salesman') and \
            not self.env['sale.order'].check_access_rights('write', raise_exception=False)

        if sudo:
            self = self.sudo()

        res = super(SaleSubscriptionCloseReasonWizard, self.with_context(close_wizard=True)).set_close()

        if sudo:
            subscription = self.env['sale.order'].browse(self.env.context.get('active_id'))
            subscription.message_post(body='Contract closed manually by %s' % user.name)

        return res

class SaleSubscriptionTemplate(models.Model):

    _inherit = 'sale.order.template'

    oe_apps_uom_id = fields.Many2one('uom.uom', 'UOM applied to Odoo Apps')
    oe_users_uom_id = fields.Many2one('uom.uom', 'UOM applied to Odoo Users')
    oe_mnt_uom_id = fields.Many2one('uom.uom', 'UOM applied to Maintenance')

    subscription_tag_ids = fields.Many2many(
        comodel_name='openerp.enterprise.subscription.tag',
        string='Subscription Tags',
        domain="['|', ('company_id', '=', False), ('company_id', '=', company_id)]"
    )

    def get_product_uom(self, product):
        self.ensure_one()

        if product.is_odoo_user:
            return self.oe_users_uom_id
        elif product.is_odoo_app:
            return self.oe_apps_uom_id
        elif product.categ_id == self.env.ref('openerp_enterprise.category_maintenance'):
            return self.oe_mnt_uom_id
        return product.uom_id


class OpenerpEnterpriseSubscriptionTag(models.Model):
    _name = 'openerp.enterprise.subscription.tag'
    _description = 'OE Subscription Tags'
    name = fields.Char(string='Subscription Tag', index='trigram', required=True)
    color = fields.Integer('Color Index')
    active = fields.Boolean(default=True, help="Set active to false to hide the Subscription Tag without removing it.")
    company_id = fields.Many2one('res.company', string='Company')
